# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Root wrapper for OpenStack services

Filters which commands a service is allowed to run as another user.

To use this with oslo, you should set the following in
oslo.conf:
rootwrap_config=/etc/oslo/rootwrap.conf

You also need to let the oslo user run oslo-rootwrap
as root in sudoers:
oslo ALL = (root) NOPASSWD: /usr/bin/oslo-rootwrap
                                /etc/oslo/rootwrap.conf *

Service packaging should deploy .filters files only on nodes where
they are needed, to avoid allowing more than is necessary.
"""

import configparser
import logging
import os
import sys
from types import ModuleType
from typing import NoReturn

from oslo_rootwrap import filters as filters_mod
from oslo_rootwrap import subprocess
from oslo_rootwrap import wrapper

resource: ModuleType | None
try:
    # This isn't available on all platforms (e.g. Windows).
    import resource
except ImportError:
    resource = None

LOG = logging.getLogger(__name__)

RC_UNAUTHORIZED = 99
RC_NOCOMMAND = 98
RC_BADCONFIG = 97
RC_NOEXECFOUND = 96
SIGNAL_BASE = 128


def _exit_error(
    execname: str, message: str, errorcode: int, log: bool = True
) -> NoReturn:
    print(f"{execname}: {message}", file=sys.stderr)
    if log:
        LOG.error(message)
    sys.exit(errorcode)


def daemon() -> None:
    main(run_daemon=True)


def main(run_daemon: bool = False) -> None:
    # Split arguments, require at least a command
    execname = sys.argv.pop(0)
    if run_daemon:
        if len(sys.argv) != 1:
            _exit_error(
                execname, "Extra arguments to daemon", RC_NOCOMMAND, log=False
            )
    else:
        if len(sys.argv) < 2:
            _exit_error(
                execname, "No command specified", RC_NOCOMMAND, log=False
            )

    configfile = sys.argv.pop(0)

    # Load configuration
    try:
        rawconfig = configparser.RawConfigParser()
        rawconfig.read(configfile)
        config = wrapper.RootwrapConfig(rawconfig)
    except ValueError as exc:
        msg = f"Incorrect value in {configfile}: {exc.args[0]}"
        _exit_error(execname, msg, RC_BADCONFIG, log=False)
    except configparser.Error:
        _exit_error(
            execname,
            f"Incorrect configuration file: {configfile}",
            RC_BADCONFIG,
            log=False,
        )

    if resource:
        # When use close_fds=True on Python 2.x, calling subprocess with
        # close_fds=True (which we do by default) can be inefficient when
        # the current fd ulimits are large, because it blindly closes
        # all fds in the range(1, $verylargenumber)

        # Lower our ulimit to a reasonable value to regain performance.
        fd_limits = resource.getrlimit(resource.RLIMIT_NOFILE)
        sensible_fd_limit = min(config.rlimit_nofile, fd_limits[0])
        if fd_limits[0] > sensible_fd_limit:
            # Close any fd beyond sensible_fd_limit prior adjusting our
            # rlimit to ensure all fds are closed
            for fd_entry in os.listdir('/proc/self/fd'):
                # NOTE(dmllr): In a previous patch revision non-numeric
                # dir entries were silently ignored which reviewers
                # didn't like. Readd exception handling when it occurs.
                fd = int(fd_entry)
                if fd >= sensible_fd_limit:
                    os.close(fd)
            # Unfortunately this inherits to our children, so allow them to
            # re-raise by passing through the hard limit unmodified
            resource.setrlimit(
                resource.RLIMIT_NOFILE, (sensible_fd_limit, fd_limits[1])
            )
            # This is set on import to the hard ulimit. if its defined we
            # already have imported it, so we need to update it to the new
            # limit.
            if (
                hasattr(subprocess, 'MAXFD')
                and subprocess.MAXFD > sensible_fd_limit
            ):
                subprocess.MAXFD = sensible_fd_limit

    if config.use_syslog:
        wrapper.setup_syslog(
            execname, config.syslog_log_facility, config.syslog_log_level
        )

    filters = wrapper.load_filters(config.filters_path)

    if run_daemon:
        # NOTE(dims): When not running as daemon, this import
        # slows us down just a bit. So moving it here so we have
        # it only when we need it.
        from oslo_rootwrap import daemon as daemon_mod

        daemon_mod.daemon_start(config, filters)
    else:
        run_one_command(execname, config, filters, sys.argv)


def run_one_command(
    execname: str,
    config: wrapper.RootwrapConfig,
    filters: list[filters_mod.CommandFilter],
    userargs: list[str],
) -> NoReturn:
    # Execute command if it matches any of the loaded filters
    try:
        obj = wrapper.start_subprocess(
            filters,
            userargs,
            exec_dirs=config.exec_dirs,
            log=config.use_syslog,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        returncode = obj.wait()
        # Fix returncode of Popen
        if returncode < 0:
            returncode = SIGNAL_BASE - returncode
        sys.exit(returncode)
    except wrapper.FilterMatchNotExecutable as exc:
        assert exc.match is not None  # narrow type
        msg = (
            f"Executable not found: {exc.match.exec_path} "
            f"(filter match = {exc.match.name})"
        )
        _exit_error(execname, msg, RC_NOEXECFOUND, log=config.use_syslog)
    except wrapper.NoFilterMatched:
        msg = "Unauthorized command: {} (no filter matched)".format(
            ' '.join(userargs)
        )
        _exit_error(execname, msg, RC_UNAUTHORIZED, log=config.use_syslog)



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Callable, Sequence
import logging
from multiprocessing import managers
import subprocess as stdlib_subprocess
import threading
import time
from typing import Any
import weakref

import oslo_rootwrap
from oslo_rootwrap import daemon
from oslo_rootwrap import jsonrpc
from oslo_rootwrap import subprocess

if oslo_rootwrap._patched_socket:
    # We have to use slow version of recvall with eventlet because of a bug in
    # GreenSocket.recv_into:
    # https://bitbucket.org/eventlet/eventlet/pull-request/41
    # This check happens here instead of jsonrpc to avoid importing eventlet
    # from daemon code that is run with root privileges.
    jsonrpc.JsonConnection.recvall = (  # type: ignore[method-assign]
        jsonrpc.JsonConnection._recvall_slow
    )

ClientManager = daemon.get_manager_class()
LOG = logging.getLogger(__name__)
SHUTDOWN_RETRIES = 3


class Client:
    _start_command: Sequence[str]
    _initialized: bool
    _need_restart: bool
    _mutex: threading.Lock
    _manager: managers.BaseManager | None
    _proxy: Any  # RootwrapClass proxy
    _process: stdlib_subprocess.Popen[bytes] | None
    _finalize: Callable[[], None] | None
    _exec_sem: threading.Lock

    def __init__(self, rootwrap_daemon_cmd: Sequence[str]) -> None:
        self._start_command = rootwrap_daemon_cmd
        self._initialized = False
        self._need_restart = False
        self._mutex = threading.Lock()
        self._manager = None
        self._proxy = None
        self._process = None
        self._finalize = None
        # This is for eventlet compatibility. multiprocessing stores
        # daemon connection in ForkAwareLocal, so this won't be
        # needed with the threading module.
        self._exec_sem = threading.Lock()

    def _initialize(self) -> None:
        if self._process is not None and self._process.poll() is not None:
            LOG.warning(
                "Leaving behind already spawned process with pid %d, "
                "root should kill it if it's still there (I can't)",
                self._process.pid,
            )

        process_obj = subprocess.Popen(
            self._start_command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
        )
        LOG.debug(
            "Popen for %s command has been instantiated", self._start_command
        )

        self._process = process_obj
        assert process_obj.stdout is not None  # narrow type
        assert process_obj.stderr is not None  # narrow type
        socket_path = process_obj.stdout.readline()[:-1].decode('utf-8')
        authkey = process_obj.stdout.read(32)
        if process_obj.poll() is not None:
            stderr = process_obj.stderr.read()
            # NOTE(yorik-sar): don't expose stdout here
            raise Exception(
                f"Failed to spawn rootwrap process.\nstderr:\n{stderr!r}"
            )
        LOG.info(
            "Spawned new rootwrap daemon process with pid=%d", process_obj.pid
        )

        def wait_process() -> None:
            return_code = process_obj.wait()
            LOG.info(
                "Rootwrap daemon process exit with status: %d", return_code
            )

        reap_process = threading.Thread(target=wait_process)
        reap_process.daemon = True
        reap_process.start()
        self._manager = ClientManager(socket_path, authkey)
        self._manager.connect()
        self._proxy = self._manager.rootwrap()  # type: ignore[attr-defined]
        self._finalize = weakref.finalize(
            self, self._shutdown, self._process, self._manager
        )
        self._initialized = True

    @staticmethod
    def _shutdown(
        process: stdlib_subprocess.Popen[bytes],
        manager: managers.BaseManager,
        JsonClient: type[jsonrpc.JsonClient] = jsonrpc.JsonClient,
    ) -> None:
        # Storing JsonClient in arguments because globals are set to None
        # before executing atexit routines in Python 2.x
        if process.poll() is None:
            LOG.info(
                'Stopping rootwrap daemon process with pid=%s', process.pid
            )
            for _ in range(SHUTDOWN_RETRIES):
                try:
                    manager.rootwrap().shutdown()  # type: ignore[attr-defined]
                    break
                except (EOFError, OSError):
                    break  # assume it is dead already
                except RuntimeError:
                    time.sleep(0.2)
            # We might want to wait for process to exit or kill it, but we
            # can't provide sane timeout on 2.x and we most likely don't have
            # permisions to do so
        # Invalidate manager's state so that proxy won't try to do decref
        manager._state.value = managers.State.SHUTDOWN  # type: ignore[attr-defined]

    def _ensure_initialized(self) -> None:
        with self._mutex:
            if not self._initialized:
                self._initialize()

    def _restart(self, proxy: Any) -> Any:
        with self._mutex:
            if not self._initialized:
                raise AssertionError("Client should be initialized.")
            # Verify if someone has already restarted this.
            if self._proxy is proxy:
                assert self._finalize is not None  # narrow type
                self._finalize()
                self._manager = None
                self._proxy = None
                self._initialized = False
                self._initialize()
                self._need_restart = False
            return self._proxy

    def _run_one_command(
        self, proxy: Any, cmd: list[str], stdin: str | None
    ) -> tuple[int, str, str]:
        """Wrap proxy.run_one_command, setting _need_restart on an exception.

        Usually it should be enough to drain stale data on socket
        rather than to restart, but we cannot do draining easily.
        """
        try:
            _need_restart = True
            res: tuple[int, str, str] = proxy.run_one_command(cmd, stdin)
            _need_restart = False
            return res
        finally:
            if _need_restart:
                self._need_restart = True

    def execute(
        self, cmd: list[str], stdin: str | None = None
    ) -> tuple[int, str, str]:
        with self._exec_sem:
            self._ensure_initialized()
            proxy = self._proxy
            retry = False
            if self._need_restart:
                proxy = self._restart(proxy)
            try:
                res = self._run_one_command(proxy, cmd, stdin)
            except (EOFError, OSError):
                retry = True
            # res can be None if we received final None sent by dying
            # server thread instead of response to our
            # request. Process is most likely to be dead at this
            # point.
            if retry or res is None:
                proxy = self._restart(proxy)
                res = self._run_one_command(proxy, cmd, stdin)
        return res



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import io
import logging
import os
import shutil
import signal
import sys
import threading
import time
from unittest import mock

try:
    import eventlet
except ImportError:
    eventlet = None

import fixtures
from testtools import content

from oslo_rootwrap import client
from oslo_rootwrap import cmd
from oslo_rootwrap import subprocess
from oslo_rootwrap.tests import functional_base
from oslo_rootwrap.tests import run_daemon


class RootwrapTest(functional_base._FunctionalBase):
    def setUp(self):
        super().setUp()
        self.cmd = [
            sys.executable,
            '-c',
            'from oslo_rootwrap import cmd; cmd.main()',
            self.config_file,
        ]

    def execute(self, cmd, stdin=None):
        proc = subprocess.Popen(
            self.cmd + cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = proc.communicate(stdin)
        self.addDetail(
            'stdout', content.text_content(out.decode('utf-8', 'replace'))
        )
        self.addDetail(
            'stderr', content.text_content(err.decode('utf-8', 'replace'))
        )
        return proc.returncode, out, err

    def test_run_once(self):
        self._test_run_once(expect_byte=True)

    def test_run_with_stdin(self):
        self._test_run_with_stdin(expect_byte=True)


class RootwrapDaemonTest(functional_base._FunctionalBase):
    def assert_unpatched(self):
        # We need to verify that these tests are run without eventlet patching
        if eventlet and eventlet.patcher.is_monkey_patched('socket'):
            self.fail(
                "Standard library should not be patched by eventlet"
                " for this test"
            )

    def setUp(self):
        self.assert_unpatched()

        super().setUp()

        # Collect daemon logs
        daemon_log = io.BytesIO()
        p = mock.patch(
            'oslo_rootwrap.subprocess.Popen',
            run_daemon.forwarding_popen(daemon_log),
        )
        p.start()
        self.addCleanup(p.stop)

        # Collect client logs
        client_log = io.StringIO()
        handler = logging.StreamHandler(client_log)
        log_format = run_daemon.log_format.replace('+', ' ')
        handler.setFormatter(logging.Formatter(log_format))
        logger = logging.getLogger('oslo_rootwrap')
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        self.addCleanup(logger.removeHandler, handler)

        # Add all logs as details
        @self.addCleanup
        def add_logs():
            self.addDetail(
                'daemon_log',
                content.Content(
                    content.UTF8_TEXT, lambda: [daemon_log.getvalue()]
                ),
            )
            self.addDetail(
                'client_log',
                content.Content(
                    content.UTF8_TEXT,
                    lambda: [client_log.getvalue().encode('utf-8')],
                ),
            )

        # Create client
        self.client = client.Client(
            [sys.executable, run_daemon.__file__, self.config_file]
        )

        # _finalize is set during Client.execute()
        @self.addCleanup
        def finalize_client():
            if self.client._initialized:
                assert self.client._finalize is not None  # narrow type
                self.client._finalize()

        self.execute = self.client.execute

    def test_run_once(self):
        self._test_run_once(expect_byte=False)

    def test_run_with_stdin(self):
        self._test_run_with_stdin(expect_byte=False)

    def test_run_with_later_install_cmd(self):
        code, out, err = self.execute(['later_install_cmd'])
        self.assertEqual(cmd.RC_NOEXECFOUND, code)
        # Install cmd and try again
        shutil.copy('/bin/echo', self.later_cmd)
        code, out, err = self.execute(['later_install_cmd'])
        # Expect successfully run the cmd
        self.assertEqual(0, code)

    def test_daemon_ressurection(self):
        # Let the client start a daemon
        self.execute(['cat'])
        # Make daemon go away
        assert self.client._process is not None
        os.kill(self.client._process.pid, signal.SIGTERM)
        # Expect client to successfully restart daemon and run simple request
        self.test_run_once()

    def test_daemon_timeout(self):
        # Let the client start a daemon
        self.execute(['echo'])
        # Make daemon timeout
        with mock.patch.object(self.client, '_restart') as restart:
            time.sleep(15)
            self.execute(['echo'])
            restart.assert_called_once()

    def _exec_thread(self, fifo_path: str) -> None:
        try:
            # Run a shell script that signals calling process through FIFO and
            # then hangs around for 1 sec
            self._thread_res: (
                tuple[int, str | bytes, str | bytes] | Exception
            ) = self.execute(
                ['sh', '-c', f'echo > "{fifo_path}"; sleep 1; echo OK']
            )
        except Exception as e:
            self._thread_res = e

    def test_graceful_death(self):
        # Create a fifo in a temporary dir
        tmpdir = self.useFixture(fixtures.TempDir()).path
        fifo_path = os.path.join(tmpdir, 'fifo')
        os.mkfifo(fifo_path)
        # Start daemon
        self.execute(['cat'])
        # Begin executing shell script
        t = threading.Thread(target=self._exec_thread, args=(fifo_path,))
        t.start()
        # Wait for shell script to actually start
        with open(fifo_path) as f:
            f.readline()
        # Gracefully kill daemon process
        assert self.client._process is not None
        os.kill(self.client._process.pid, signal.SIGTERM)
        # Expect daemon to wait for our request to finish
        t.join()
        if isinstance(self._thread_res, Exception):
            raise self._thread_res  # Python 3 will even provide nice traceback
        code, out, err = self._thread_res
        self.assertEqual(0, code)
        self.assertEqual('OK\n', out)
        self.assertEqual('', err)

    @contextlib.contextmanager
    def _test_daemon_cleanup(self):
        # Start a daemon
        self.execute(['cat'])
        assert self.client._manager is not None
        socket_path = self.client._manager.address
        assert isinstance(socket_path, str)
        # Stop it one way or another
        yield
        process = self.client._process
        assert process is not None
        stop = threading.Event()

        # Start background thread that would kill process in 1 second if it
        # doesn't die by then
        def sleep_kill():
            stop.wait(3)
            if not stop.is_set():
                os.kill(process.pid, signal.SIGKILL)

        threading.Thread(target=sleep_kill).start()
        # Wait for process to finish one way or another
        process.wait()
        # Notify background thread that process is dead (no need to kill it)
        stop.set()
        # Fail if the process got killed by the background thread
        self.assertNotEqual(
            -signal.SIGKILL,
            process.returncode,
            "Server haven't stopped in one second",
        )
        # Verify that socket is deleted
        self.assertFalse(
            os.path.exists(socket_path),
            "Server didn't remove its temporary directory",
        )

    def test_daemon_cleanup_client(self):
        # Run _test_daemon_cleanup stopping daemon as Client instance would
        # normally do
        with self._test_daemon_cleanup():
            assert self.client._finalize is not None  # narrow type
            self.client._finalize()

    def test_daemon_cleanup_signal(self):
        # Run _test_daemon_cleanup stopping daemon with SIGTERM signal
        with self._test_daemon_cleanup():
            assert self.client._process is not None
            os.kill(self.client._process.pid, signal.SIGTERM)



# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections.abc import Sequence
import os
import re
import shutil
import sys

NETNS_VARS = ('net', 'netn', 'netns')
EXEC_VARS = ('e', 'ex', 'exe', 'exec')


if sys.platform != 'win32':
    # NOTE(claudiub): pwd is a Linux-specific library, and currently there is
    # no Windows support for oslo.rootwrap.
    import pwd


def _getuid(user: str) -> int:
    """Return uid for user."""
    return pwd.getpwnam(user).pw_uid


def realpath(path: str) -> str:
    """Return the real absolute path.

    If the execution directory does not exist, os.getcwd() raises a
    FileNotFoundError exception. In this case, unset the exception and return
    an empty string.
    """
    try:
        return os.path.realpath(path)
    except FileNotFoundError:
        return ''


class CommandFilter:
    """Command filter only checking that the 1st argument matches exec_path."""

    def __init__(self, exec_path: str, run_as: str, *args: str) -> None:
        self.name = ''
        self.exec_path = exec_path
        self.run_as = run_as
        self.args = args
        self.real_exec: str | None = None

    def get_exec(self, exec_dirs: list[str] | None = None) -> str | None:
        """Returns existing executable, or empty string if none found."""
        exec_dirs = exec_dirs or []
        if self.real_exec is not None:
            return self.real_exec
        if os.path.isabs(self.exec_path):
            if os.access(self.exec_path, os.X_OK):
                self.real_exec = self.exec_path
        else:
            for binary_path in exec_dirs:
                expanded_path = os.path.join(binary_path, self.exec_path)
                if os.access(expanded_path, os.X_OK):
                    self.real_exec = expanded_path
                    break
        return self.real_exec

    def match(self, userargs: list[str] | None) -> bool:
        """Only check that the first argument (command) matches exec_path."""
        if userargs:
            base_path_matches = os.path.basename(self.exec_path) == userargs[0]
            exact_path_matches = self.exec_path == userargs[0]
            return exact_path_matches or base_path_matches
        return False

    def preexec(self) -> None:
        """Setuid in subprocess right before command is invoked."""
        if self.run_as != 'root':
            os.setuid(_getuid(self.run_as))

    def get_command(
        self, userargs: list[str], exec_dirs: list[str] | None = None
    ) -> list[str]:
        """Returns command to execute."""
        exec_dirs = exec_dirs or []
        to_exec = self.get_exec(exec_dirs=exec_dirs) or self.exec_path
        return [to_exec] + list(userargs[1:])

    def get_environment(self, userargs: list[str]) -> dict[str, str] | None:
        """Returns specific environment to set, None if none."""
        return None


class RegExpFilter(CommandFilter):
    """Command filter doing regexp matching for every argument."""

    def match(self, userargs: list[str] | None) -> bool:
        # Early skip if command or number of args don't match
        if not userargs or len(self.args) != len(userargs):
            # DENY: argument numbers don't match
            return False
        # Compare each arg (anchoring pattern explicitly at end of string)
        for pattern, arg in zip(self.args, userargs):
            try:
                if not re.match(pattern + '$', arg):
                    # DENY: Some arguments did not match
                    return False
            except re.error:
                # DENY: Badly-formed filter
                return False
        # ALLOW: All arguments matched
        return True


class PathFilter(CommandFilter):
    """Command filter checking that path arguments are within given dirs

    One can specify the following constraints for command arguments:
        1) pass     - pass an argument as is to the resulting command
        2) some_str - check if an argument is equal to the given string
        3) abs path - check if a path argument is within the given base dir

    A typical rootwrapper filter entry looks like this:
        # cmdname: filter name, raw command, user, arg_i_constraint [, ...]
        chown: PathFilter, /bin/chown, root, nova, /var/lib/images

    """

    def match(self, userargs: list[str] | None) -> bool:
        if not userargs or len(userargs) < 2:
            return False

        arguments = userargs[1:]

        equal_args_num = len(self.args) == len(arguments)
        exec_is_valid = super().match(userargs)
        args_equal_or_pass = all(
            arg == 'pass' or arg == value
            for arg, value in zip(self.args, arguments)
            if not os.path.isabs(arg)  # arguments not specifying abs paths
        )
        paths_are_within_base_dirs = all(
            os.path.commonprefix([arg, realpath(value)]) == arg
            for arg, value in zip(self.args, arguments)
            if os.path.isabs(arg)  # arguments specifying abs paths
        )

        return (
            equal_args_num
            and exec_is_valid
            and args_equal_or_pass
            and paths_are_within_base_dirs
        )

    def get_command(
        self, userargs: list[str], exec_dirs: list[str] | None = None
    ) -> list[str]:
        exec_dirs = exec_dirs or []
        command, arguments = userargs[0], userargs[1:]

        # convert path values to canonical ones; copy other args as is
        args = [
            realpath(value) if os.path.isabs(arg) else value
            for arg, value in zip(self.args, arguments)
        ]

        return super().get_command([command] + args, exec_dirs)


class KillFilter(CommandFilter):
    """Specific filter for the kill calls.

    1st argument is the user to run /bin/kill under
    2nd argument is the location of the affected executable
        if the argument is not absolute, it is checked against $PATH
    Subsequent arguments list the accepted signals (if any)

    This filter relies on /proc to accurately determine affected
    executable, so it will only work on procfs-capable systems (not OSX).
    """

    def __init__(self, *args: str) -> None:
        super().__init__("/bin/kill", *args)

    def _program(self, pid: int) -> str | None:
        """Determine the program associated with pid"""

        try:
            command = os.readlink(f"/proc/{pid}/exe")
        except (ValueError, OSError):
            # Incorrect PID
            return None

        # NOTE(yufang521247): /proc/PID/exe may have '\0' on the
        # end (ex: if an executable is updated or deleted), because python
        # doesn't stop at '\0' when read the target path.
        command = command.partition('\0')[0]

        # NOTE(dprince): /proc/PID/exe may have ' (deleted)' on
        # the end if an executable is updated or deleted
        if command.endswith(" (deleted)"):
            command = command[: -len(" (deleted)")]

        if os.path.isfile(command):
            return command

        # /proc/PID/exe may have been renamed with
        # a ';......' or '.#prelink#......' suffix etc.
        # So defer to /proc/PID/cmdline in that case.
        try:
            with open(f"/proc/{pid}/cmdline") as pfile:
                cmdline = pfile.read().partition('\0')[0]

            cmdline = shutil.which(cmdline) or cmdline
            if os.path.isfile(cmdline):
                command = cmdline

            # Note we don't return None if cmdline doesn't exist
            # as that will allow killing a process where the exe
            # has been removed from the system rather than updated.
            return command
        except OSError:
            return None

    def match(self, userargs: list[str] | None) -> bool:
        if not userargs or userargs[0] != "kill":
            return False
        args = list(userargs)
        if len(args) == 3:
            # A specific signal is requested
            signal = args.pop(1)
            if signal not in self.args[1:]:
                # Requested signal not in accepted list
                return False
        else:
            if len(args) != 2:
                # Incorrect number of arguments
                return False
            if len(self.args) > 1:
                # No signal requested, but filter requires specific signal
                return False

        try:
            pid = int(args[1])
        except ValueError:
            return False
        command = self._program(pid)
        if not command:
            return False

        kill_command = self.args[0]

        if os.path.isabs(kill_command):
            return kill_command == command

        return (
            os.path.isabs(command)
            and kill_command == os.path.basename(command)
            and os.path.dirname(command)
            in os.environ.get('PATH', '').split(':')
        )


class ReadFileFilter(CommandFilter):
    """Specific filter for the utils.read_file_as_root call."""

    def __init__(self, file_path: str, *args: str) -> None:
        self.file_path = file_path
        super().__init__("/bin/cat", "root", *args)

    def match(self, userargs: list[str] | None) -> bool:
        return userargs == ['cat', self.file_path]


class IpFilter(CommandFilter):
    """Specific filter for the ip utility to that does not match exec."""

    def match(self, userargs: list[str] | None) -> bool:
        if not userargs:
            return False
        if userargs[0] == 'ip':
            # Avoid the 'netns exec' command here
            for a, b in zip(userargs[1:], userargs[2:]):
                if a in NETNS_VARS:
                    return b not in EXEC_VARS
            else:
                return True
        return False


class EnvFilter(CommandFilter):
    """Specific filter for the env utility.

    Behaves like CommandFilter, except that it handles
    leading env A=B.. strings appropriately.
    """

    def _extract_env(self, arglist: Sequence[str]) -> set[str]:
        """Extract all leading NAME=VALUE arguments from arglist."""

        envs = set()
        for arg in arglist:
            if '=' not in arg:
                break
            envs.add(arg.partition('=')[0])
        return envs

    def __init__(self, exec_path: str, run_as: str, *args: str) -> None:
        super().__init__(exec_path, run_as, *args)

        env_list = self._extract_env(self.args)
        # Set exec_path to X when args are in the form of
        # env A=a B=b C=c X Y Z
        if "env" in exec_path and len(env_list) < len(self.args):
            self.exec_path = self.args[len(env_list)]

    def match(self, userargs: list[str] | None) -> bool:
        if not userargs:
            return False
        # ignore leading 'env'
        if userargs[0] == 'env':
            userargs.pop(0)

        # require one additional argument after configured ones
        if len(userargs) < len(self.args):
            return False

        # extract all env args
        user_envs = self._extract_env(userargs)
        filter_envs = self._extract_env(self.args)
        user_command = userargs[len(user_envs) : len(user_envs) + 1]

        # match first non-env argument with CommandFilter
        return bool(
            super().match(user_command)
            and len(filter_envs)
            and user_envs == filter_envs
        )

    def exec_args(self, userargs: list[str]) -> list[str]:
        args = userargs[:]

        # ignore leading 'env'
        if args[0] == 'env':
            args.pop(0)

        # Throw away leading NAME=VALUE arguments
        while args and '=' in args[0]:
            args.pop(0)

        return args

    def get_command(
        self, userargs: list[str], exec_dirs: list[str] | None = None
    ) -> list[str]:
        to_exec = self.get_exec(exec_dirs=exec_dirs or []) or self.exec_path
        return [to_exec] + self.exec_args(userargs)[1:]

    def get_environment(self, userargs: list[str]) -> dict[str, str] | None:
        env = os.environ.copy()

        # ignore leading 'env'
        if userargs[0] == 'env':
            userargs.pop(0)

        # Handle leading NAME=VALUE pairs
        for a in userargs:
            env_name, equals, env_value = a.partition('=')
            if not equals:
                break
            if env_name and env_value:
                env[env_name] = env_value

        return env


class ChainingFilter(CommandFilter):
    def exec_args(self, userargs: list[str]) -> list[str]:
        return []


class IpNetnsExecFilter(ChainingFilter):
    """Specific filter for the ip utility to that does match exec."""

    def match(self, userargs: list[str] | None) -> bool:
        # Network namespaces currently require root
        # require <ns> argument
        if not userargs or self.run_as != "root" or len(userargs) < 4:
            return False

        return (
            userargs[0] == 'ip'
            and userargs[1] in NETNS_VARS
            and userargs[2] in EXEC_VARS
        )

    def exec_args(self, userargs: list[str]) -> list[str]:
        args = userargs[4:]
        if args:
            args[0] = os.path.basename(args[0])
        return args


class ChainingRegExpFilter(ChainingFilter):
    """Command filter doing regexp matching for prefix commands.

    Remaining arguments are filtered again. This means that the command
    specified as the arguments must be also allowed to execute directly.
    """

    def match(self, userargs: list[str] | None) -> bool:
        # Early skip if number of args is smaller than the filter
        if not userargs or len(self.args) > len(userargs):
            return False
        # Compare each arg (anchoring pattern explicitly at end of string)
        for pattern, arg in zip(self.args, userargs):
            try:
                if not re.match(pattern + '$', arg):
                    # DENY: Some arguments did not match
                    return False
            except re.error:
                # DENY: Badly-formed filter
                return False
        # ALLOW: All arguments matched
        return True

    def exec_args(self, userargs: list[str]) -> list[str]:
        args = userargs[len(self.args) :]
        if args:
            args[0] = os.path.basename(args[0])
        return args



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import atexit
import math
import os
import subprocess
import sys
import timeit

from oslo_rootwrap import client

config_path = "rootwrap.conf"
num_iterations = 100


def run_plain(cmd):
    obj = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = obj.communicate()
    out = os.fsdecode(out)
    err = os.fsdecode(err)
    return obj.returncode, out, err


def run_sudo(cmd):
    return run_plain(["sudo"] + cmd)


def run_rootwrap(cmd):
    return run_plain(
        [
            "sudo",
            sys.executable,
            "-c",
            "from oslo_rootwrap import cmd; cmd.main()",
            config_path,
        ]
        + cmd
    )


run_daemon = client.Client(
    [
        "sudo",
        sys.executable,
        "-c",
        "from oslo_rootwrap import cmd; cmd.daemon()",
        config_path,
    ]
).execute


def run_one(runner, cmd):
    def __inner():
        code, out, err = runner(cmd)
        assert err == "", "Stderr not empty:\n" + err
        assert code == 0, "Command failed"

    return __inner


runners = [
    ("{0}", run_plain),
    ("sudo {0}", run_sudo),
    ("sudo rootwrap conf {0}", run_rootwrap),
    ("daemon.run('{0}')", run_daemon),
]


def get_time_string(sec):
    if sec > 0.9:
        return f"{sec:7.3f}s "
    elif sec > 0.0009:
        return f"{sec * 1000.0:7.3f}ms"
    else:
        return f"{sec * 1000000.0:7.3f}us"


def run_bench(cmd, runners):
    strcmd = ' '.join(cmd)
    max_name_len = max(len(name) for name, _ in runners) + len(strcmd) - 3
    print(f"Running '{strcmd}':")
    print(
        "{0:^{1}} :".format("method", max_name_len),
        "".join(map("{:^10}".format, ["min", "avg", "max", "dev"])),
    )
    for name, runner in runners:
        results = timeit.repeat(
            run_one(runner, cmd), repeat=num_iterations, number=1
        )
        avg = sum(results) / num_iterations
        min_ = min(results)
        max_ = max(results)
        dev = math.sqrt(sum((r - avg) ** 2 for r in results) / num_iterations)
        print(
            "{0:>{1}} :".format(name.format(strcmd), max_name_len),
            " ".join(map(get_time_string, [min_, avg, max_, dev])),
        )


def main():
    os.chdir(os.path.dirname(__file__))
    code, _, _ = run_sudo(["-vn"])
    if code:
        print("We need you to authorize with sudo to run this benchmark")
        run_sudo(["-v"])

    run_bench(["ip", "a"], runners)
    run_sudo(["ip", "netns", "add", "bench_ns"])
    atexit.register(run_sudo, ["ip", "netns", "delete", "bench_ns"])
    run_bench('ip netns exec bench_ns ip a'.split(), runners[1:])


if __name__ == "__main__":
    main()



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import logging
from multiprocessing import managers
import os
import shutil
import signal
import stat
import sys
import tempfile
import threading
import time
from types import FrameType

from oslo_rootwrap import cmd
from oslo_rootwrap import filters as filters_mod
from oslo_rootwrap import jsonrpc
from oslo_rootwrap import subprocess
from oslo_rootwrap import wrapper

LOG = logging.getLogger(__name__)

# Since multiprocessing supports only pickle and xmlrpclib for serialization of
# RPC requests and responses, we declare another 'jsonrpc' serializer

managers.listener_client['jsonrpc'] = (  # type: ignore[attr-defined]
    jsonrpc.JsonListener,
    jsonrpc.JsonClient,
)


class RootwrapClass:
    last_called: float
    daemon_timeout: int
    timeout: threading.Timer

    def __init__(
        self,
        config: wrapper.RootwrapConfig,
        filters: list[filters_mod.CommandFilter],
    ) -> None:
        self.config = config
        self.filters = filters
        self.reset_timer()
        self.prepare_timer(config)

    def run_one_command(
        self, userargs: list[str], stdin: str | None = None
    ) -> tuple[int, str, str]:
        self.reset_timer()
        try:
            obj = wrapper.start_subprocess(
                self.filters,
                userargs,
                exec_dirs=self.config.exec_dirs,
                log=self.config.use_syslog,
                close_fds=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except wrapper.FilterMatchNotExecutable:
            LOG.warning("Executable not found for: %s", ' '.join(userargs))
            return cmd.RC_NOEXECFOUND, "", ""

        except wrapper.NoFilterMatched:
            LOG.warning(
                "Unauthorized command: %s (no filter matched)",
                ' '.join(userargs),
            )
            return cmd.RC_UNAUTHORIZED, "", ""

        stdin_bytes: bytes | None = None
        if stdin is not None:
            stdin_bytes = os.fsencode(stdin)
        out, err = obj.communicate(stdin_bytes)
        out_str = os.fsdecode(out)
        err_str = os.fsdecode(err)
        return obj.returncode or 0, out_str, err_str

    @classmethod
    def reset_timer(cls) -> None:
        cls.last_called = time.time()

    @classmethod
    def cancel_timer(cls) -> None:
        try:
            cls.timeout.cancel()
        except RuntimeError:
            pass

    @classmethod
    def prepare_timer(
        cls, config: wrapper.RootwrapConfig | None = None
    ) -> None:
        if config is not None:
            cls.daemon_timeout = config.daemon_timeout
        # Wait a bit longer to avoid rounding errors
        timeout = (
            max(cls.last_called + cls.daemon_timeout - time.time(), 0) + 1
        )
        if getattr(cls, 'timeout', None):
            # Another timer is already initialized
            return
        cls.timeout = threading.Timer(timeout, cls.handle_timeout)
        cls.timeout.start()

    @classmethod
    def handle_timeout(cls) -> None:
        if cls.last_called < time.time() - cls.daemon_timeout:
            cls.shutdown()

        cls.prepare_timer()

    @staticmethod
    def shutdown() -> None:
        # Suicide to force break of the main thread
        os.kill(os.getpid(), signal.SIGINT)


def get_manager_class(
    config: wrapper.RootwrapConfig | None = None,
    filters: list[filters_mod.CommandFilter] | None = None,
) -> type[managers.BaseManager]:
    class RootwrapManager(managers.BaseManager):
        def __init__(
            self, address: str | None = None, authkey: bytes | None = None
        ) -> None:
            # Force jsonrpc because neither pickle nor xmlrpclib is secure
            super().__init__(address, authkey, serializer='jsonrpc')

    if config is not None:
        assert filters is not None  # narrow type
        partial_class = functools.partial(RootwrapClass, config, filters)
        RootwrapManager.register('rootwrap', partial_class)
    else:
        RootwrapManager.register('rootwrap')

    return RootwrapManager


def daemon_start(
    config: wrapper.RootwrapConfig,
    filters: list[filters_mod.CommandFilter],
) -> None:
    temp_dir = tempfile.mkdtemp(prefix='rootwrap-')
    LOG.debug("Created temporary directory %s", temp_dir)
    try:
        # allow everybody to find the socket
        rwxr_xr_x = (
            stat.S_IRWXU
            | stat.S_IRGRP
            | stat.S_IXGRP
            | stat.S_IROTH
            | stat.S_IXOTH
        )
        os.chmod(temp_dir, rwxr_xr_x)
        socket_path = os.path.join(temp_dir, "rootwrap.sock")
        LOG.debug("Will listen on socket %s", socket_path)
        manager_cls = get_manager_class(config, filters)
        manager = manager_cls(address=socket_path)
        server = manager.get_server()
        try:
            # allow everybody to connect to the socket
            rw_rw_rw_ = (
                stat.S_IRUSR
                | stat.S_IWUSR
                | stat.S_IRGRP
                | stat.S_IWGRP
                | stat.S_IROTH
                | stat.S_IWOTH
            )
            os.chmod(socket_path, rw_rw_rw_)
            sys.stdout.buffer.write(socket_path.encode('utf-8'))
            sys.stdout.buffer.write(b'\n')
            # type stubs don't expose this paramter currently
            sys.stdout.buffer.write(bytes(server.authkey))  # type: ignore[attr-defined]
            sys.stdin.close()
            sys.stdout.close()
            sys.stderr.close()
            # Gracefully shutdown on INT or TERM signals
            stop = functools.partial(daemon_stop, server)
            signal.signal(signal.SIGTERM, stop)
            signal.signal(signal.SIGINT, stop)
            LOG.info("Starting rootwrap daemon main loop")
            server.serve_forever()
        finally:
            conn: jsonrpc.JsonListener = server.listener  # type: ignore[attr-defined]
            # This will break accept() loop with EOFError if it was not in the
            # main thread (as in Python 3.x)
            conn.close()
            # Closing all currently connected client sockets for reading to
            # break worker threads blocked on recv()
            for cl_conn in conn.get_accepted():
                try:
                    cl_conn.half_close()
                except Exception:
                    # Most likely the socket have already been closed
                    LOG.debug("Failed to close connection")
            RootwrapClass.cancel_timer()
            LOG.info("Waiting for all client threads to finish.")
            for thread in threading.enumerate():
                if thread.daemon:
                    LOG.debug("Joining thread %s", thread)
                    thread.join()
    finally:
        LOG.debug("Removing temporary directory %s", temp_dir)
        shutil.rmtree(temp_dir)


def daemon_stop(
    server: managers.Server, signum: int, frame: FrameType | None
) -> None:
    LOG.info("Got signal %s. Shutting down server", signum)
    # Signals are caught in the main thread which means this handler will run
    # in the middle of serve_forever() loop. It will catch this exception and
    # properly return. Since all threads created by server_forever are
    # daemonic, we need to join them afterwards. In Python 3 we can just hit
    # stop_event instead.
    server.stop_event.set()  # type: ignore[attr-defined]



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import pwd

try:
    import eventlet
except ImportError:
    eventlet = None

import fixtures
import testtools

from oslo_rootwrap import cmd


class _FunctionalBase(testtools.TestCase):
    def setUp(self) -> None:
        super().setUp()
        tmpdir = self.useFixture(fixtures.TempDir()).path
        self.config_file = os.path.join(tmpdir, 'rootwrap.conf')
        self.later_cmd = os.path.join(tmpdir, 'later_install_cmd')
        filters_dir = os.path.join(tmpdir, 'filters.d')
        filters_file = os.path.join(tmpdir, 'filters.d', 'test.filters')
        os.mkdir(filters_dir)
        with open(self.config_file, 'w') as f:
            f.write(f"""[DEFAULT]
filters_path={filters_dir}
daemon_timeout=10
exec_dirs=/bin""")
        with open(filters_file, 'w') as f:
            f.write(f"""[Filters]
echo: CommandFilter, /bin/echo, root
cat: CommandFilter, /bin/cat, root
sh: CommandFilter, /bin/sh, root
id: CommandFilter, /usr/bin/id, nobody
unknown_cmd: CommandFilter, /unknown/unknown_cmd, root
later_install_cmd: CommandFilter, {self.later_cmd}, root
""")

    def _test_run_once(self, expect_byte: bool = True) -> None:
        code, out, err = self.execute(['echo', 'teststr'])
        self.assertEqual(0, code)
        expect_out: str | bytes
        expect_err: str | bytes
        if expect_byte:
            expect_out = b'teststr\n'
            expect_err = b''
        else:
            expect_out = 'teststr\n'
            expect_err = ''
        self.assertEqual(expect_out, out)
        self.assertEqual(expect_err, err)

    def _test_run_with_stdin(self, expect_byte: bool = True) -> None:
        code, out, err = self.execute(['cat'], stdin=b'teststr')
        self.assertEqual(0, code)
        expect_out: str | bytes
        expect_err: str | bytes
        if expect_byte:
            expect_out = b'teststr'
            expect_err = b''
        else:
            expect_out = 'teststr'
            expect_err = ''
        self.assertEqual(expect_out, out)
        self.assertEqual(expect_err, err)

    def test_run_with_path(self):
        code, out, err = self.execute(['/bin/echo', 'teststr'])
        self.assertEqual(0, code)

    def test_run_with_bogus_path(self):
        code, out, err = self.execute(['/home/bob/bin/echo', 'teststr'])
        self.assertEqual(cmd.RC_UNAUTHORIZED, code)

    def test_run_command_not_found(self):
        code, out, err = self.execute(['unknown_cmd'])
        self.assertEqual(cmd.RC_NOEXECFOUND, code)

    def test_run_unauthorized_command(self):
        code, out, err = self.execute(['unauthorized_cmd'])
        self.assertEqual(cmd.RC_UNAUTHORIZED, code)

    def test_run_as(self):
        if os.getuid() != 0:
            self.skip('Test requires root (for setuid)')

        # Should run as 'nobody'
        code, out, err = self.execute(['id', '-u'])
        self.assertEqual(pwd.getpwnam('nobody').pw_uid, int(out.strip()))

        # Should run as 'root'
        code, out, err = self.execute(['sh', '-c', 'id -u'])
        self.assertEqual(0, int(out.strip()))



# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import configparser
import logging
import logging.handlers
import os
import signal
import sys
from typing import Any

from oslo_rootwrap import filters
from oslo_rootwrap import subprocess

if sys.platform != 'win32':
    import pwd

LOG = logging.getLogger(__name__)


class NoFilterMatched(Exception):
    """This exception is raised when no filter matched."""

    pass


class FilterMatchNotExecutable(Exception):
    """Raised when a filter matched but no executable was found."""

    def __init__(
        self, match: filters.CommandFilter | None = None, **kwargs: Any
    ) -> None:
        self.match = match


class RootwrapConfig:
    filters_path: list[str]
    exec_dirs: list[str]
    syslog_log_facility: int
    syslog_log_level: int
    use_syslog: bool
    daemon_timeout: int
    rlimit_nofile: int

    def __init__(self, config: configparser.RawConfigParser) -> None:
        # filters_path
        self.filters_path = config.get("DEFAULT", "filters_path").split(",")

        # exec_dirs
        if config.has_option("DEFAULT", "exec_dirs"):
            self.exec_dirs = config.get("DEFAULT", "exec_dirs").split(",")
        else:
            self.exec_dirs = []
            # Use system PATH if exec_dirs is not specified
            if "PATH" in os.environ:
                self.exec_dirs = os.environ['PATH'].split(':')

        # syslog_log_facility
        if config.has_option("DEFAULT", "syslog_log_facility"):
            v = config.get("DEFAULT", "syslog_log_facility")
            facility_names = logging.handlers.SysLogHandler.facility_names
            facility: int | None = getattr(
                logging.handlers.SysLogHandler, v, None
            )
            if facility is None and v in facility_names:
                facility = facility_names.get(v)
            if facility is None:
                raise ValueError(f'Unexpected syslog_log_facility: {v}')
            self.syslog_log_facility = facility
        else:
            self.syslog_log_facility = (
                logging.handlers.SysLogHandler.LOG_SYSLOG
            )

        # syslog_log_level
        if config.has_option("DEFAULT", "syslog_log_level"):
            v = config.get("DEFAULT", "syslog_log_level")
            level_name = v.upper()
            level: int | str = logging.getLevelName(level_name)
            if isinstance(level, str):
                raise ValueError(f'Unexpected syslog_log_level: {v!r}')
            self.syslog_log_level = level
        else:
            self.syslog_log_level = logging.ERROR

        # use_syslog
        if config.has_option("DEFAULT", "use_syslog"):
            self.use_syslog = config.getboolean("DEFAULT", "use_syslog")
        else:
            self.use_syslog = False

        # daemon_timeout
        if config.has_option("DEFAULT", "daemon_timeout"):
            self.daemon_timeout = int(config.get("DEFAULT", "daemon_timeout"))
        else:
            self.daemon_timeout = 600

        # fd ulimit
        if config.has_option("DEFAULT", "rlimit_nofile"):
            self.rlimit_nofile = int(config.get("DEFAULT", "rlimit_nofile"))
        else:
            self.rlimit_nofile = 1024


def setup_syslog(execname: str, facility: int, level: int) -> None:
    try:
        handler = logging.handlers.SysLogHandler(
            address='/dev/log', facility=facility
        )
    except OSError:
        LOG.warning(
            "Unable to setup syslog, maybe /dev/log socket needs "
            "to be restarted. Ignoring syslog configuration "
            "options."
        )
        return

    rootwrap_logger = logging.getLogger()
    rootwrap_logger.setLevel(level)
    handler.setFormatter(
        logging.Formatter(os.path.basename(execname) + ': %(message)s')
    )
    rootwrap_logger.addHandler(handler)


def build_filter(class_name: str, *args: Any) -> filters.CommandFilter | None:
    """Returns a filter object of class class_name."""
    match class_name:
        case "CommandFilter":
            return filters.CommandFilter(*args)
        case "RegExpFilter":
            return filters.RegExpFilter(*args)
        case "PathFilter":
            return filters.PathFilter(*args)
        case "KillFilter":
            return filters.KillFilter(*args)
        case "ReadFileFilter":
            return filters.ReadFileFilter(*args)
        case "IpFilter":
            return filters.IpFilter(*args)
        case "EnvFilter":
            return filters.EnvFilter(*args)
        case "ChainingFilter":
            return filters.ChainingFilter(*args)
        case "IpNetnsExecFilter":
            return filters.IpNetnsExecFilter(*args)
        case "ChainingRegExpFilter":
            return filters.ChainingRegExpFilter(*args)
        case _:
            LOG.warning(
                "Skipping unknown filter class (%s) specified "
                "in filter definitions",
                class_name,
            )
            return None


def load_filters(filters_path: list[str]) -> list[filters.CommandFilter]:
    """Load filters from a list of directories."""
    filterlist = []
    for filterdir in filters_path:
        if not os.path.isdir(filterdir):
            continue
        for filterfile in filter(
            lambda f: not f.startswith('.'), os.listdir(filterdir)
        ):
            filterfilepath = os.path.join(filterdir, filterfile)
            if not os.path.isfile(filterfilepath):
                continue
            filterconfig = configparser.RawConfigParser(strict=False)
            filterconfig.read(filterfilepath)
            for name, value in filterconfig.items("Filters"):
                filterdefinition = [s.strip() for s in value.split(',')]
                newfilter = build_filter(*filterdefinition)
                if newfilter is None:
                    continue
                newfilter.name = name
                filterlist.append(newfilter)
    # And always include privsep-helper
    privsep = build_filter("CommandFilter", "privsep-helper", "root")
    assert privsep is not None  # narrow type
    privsep.name = "privsep-helper"
    filterlist.append(privsep)
    return filterlist


def match_filter(
    filter_list: list[filters.CommandFilter],
    userargs: list[str],
    exec_dirs: list[str] | None = None,
) -> filters.CommandFilter:
    """Checks user command and arguments through command filters.

    Returns the first matching filter.

    Raises NoFilterMatched if no filter matched.
    Raises FilterMatchNotExecutable if no executable was found for the
    best filter match.
    """
    first_not_executable_filter = None
    exec_dirs = exec_dirs or []

    for f in filter_list:
        if f.match(userargs):
            if isinstance(f, filters.ChainingFilter):
                # This command calls exec verify that remaining args
                # matches another filter.
                def non_chain_filter(fltr: filters.CommandFilter) -> bool:
                    return fltr.run_as == f.run_as and not isinstance(
                        fltr, filters.ChainingFilter
                    )

                leaf_filters = [
                    fltr for fltr in filter_list if non_chain_filter(fltr)
                ]
                args = f.exec_args(userargs)
                if not args:
                    continue
                try:
                    match_filter(leaf_filters, args, exec_dirs=exec_dirs)
                except (NoFilterMatched, FilterMatchNotExecutable):
                    continue

            # Try other filters if executable is absent
            if not f.get_exec(exec_dirs=exec_dirs):
                if not first_not_executable_filter:
                    first_not_executable_filter = f
                continue
            # Otherwise return matching filter for execution
            return f

    if first_not_executable_filter:
        # A filter matched, but no executable was found for it
        raise FilterMatchNotExecutable(match=first_not_executable_filter)

    # No filter matched
    raise NoFilterMatched()


def _getlogin() -> str | None:
    try:
        return os.getlogin()
    except OSError:
        return (
            os.getenv('USER') or os.getenv('USERNAME') or os.getenv('LOGNAME')
        )


def start_subprocess(
    filter_list: list[filters.CommandFilter],
    userargs: list[str],
    exec_dirs: list[str] | None = None,
    log: bool = False,
    **kwargs: Any,
) -> subprocess.Popen[bytes]:
    filtermatch = match_filter(filter_list, userargs, exec_dirs or [])

    command = filtermatch.get_command(userargs, exec_dirs or [])
    if log:
        LOG.info(
            "(%s > %s) Executing %s (filter match = %s)",
            _getlogin(),
            pwd.getpwuid(os.getuid())[0],
            command,
            filtermatch.name,
        )

    def preexec() -> None:
        # Python installs a SIGPIPE handler by default. This is
        # usually not what non-Python subprocesses expect.
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        filtermatch.preexec()

    obj = subprocess.Popen(
        command,
        preexec_fn=preexec,
        env=filtermatch.get_environment(userargs),
        **kwargs,
    )
    return obj



#    Copyright 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import configparser
import logging
import logging.handlers
import os
import shutil
import tempfile
from unittest import mock
import uuid

import fixtures
import testtools

from oslo_rootwrap import cmd
from oslo_rootwrap import daemon
from oslo_rootwrap import filters
from oslo_rootwrap import subprocess
from oslo_rootwrap import wrapper


class RootwrapLoaderTestCase(testtools.TestCase):
    def test_privsep_in_loader(self):
        privsep = ["privsep-helper", "--context", "foo"]
        filterlist = wrapper.load_filters([])

        # mock out get_exec because
        with mock.patch.object(filters.CommandFilter, 'get_exec') as ge:
            ge.return_value = "/fake/privsep-helper"
            filtermatch = wrapper.match_filter(filterlist, privsep)

            self.assertIsNotNone(filtermatch)
            self.assertEqual(
                ["/fake/privsep-helper", "--context", "foo"],
                filtermatch.get_command(privsep),
            )

    def test_strict_switched_off_in_configparser(self):
        temp_dir = self.useFixture(fixtures.TempDir()).path
        os.mkdir(os.path.join(temp_dir, 'nested'))
        temp_file = os.path.join(temp_dir, 'test.conf')
        f = open(temp_file, 'w')
        f.write("""[Filters]
privsep: PathFilter, privsep-helper, root
privsep: PathFilter, privsep-helper, root
""")
        f.close()
        filterlist = wrapper.load_filters([temp_dir])
        self.assertIsNotNone(filterlist)


class RootwrapTestCase(testtools.TestCase):
    if os.path.exists('/sbin/ip'):
        _ip = '/sbin/ip'
    else:
        _ip = '/bin/ip'

    def setUp(self):
        super().setUp()
        self.filters = [
            filters.RegExpFilter("/bin/ls", "root", 'ls', '/[a-z]+'),
            filters.CommandFilter("/usr/bin/foo_bar_not_exist", "root"),
            filters.RegExpFilter("/bin/cat", "root", 'cat', '/[a-z]+'),
            filters.CommandFilter("/nonexistent/cat", "root"),
            filters.CommandFilter("/bin/cat", "root"),  # Keep this one last
        ]

    def test_CommandFilter(self):
        f = filters.CommandFilter("sleep", 'root', '10')
        self.assertFalse(f.match(["sleep2"]))

        # verify that any arguments are accepted
        self.assertTrue(f.match(["sleep"]))
        self.assertTrue(f.match(["sleep", "anything"]))
        self.assertTrue(f.match(["sleep", "10"]))
        f = filters.CommandFilter("sleep", 'root')
        self.assertTrue(f.match(["sleep", "10"]))

    def test_empty_commandfilter(self):
        f = filters.CommandFilter("sleep", "root")
        self.assertFalse(f.match([]))
        self.assertFalse(f.match(None))

    def test_empty_regexpfilter(self):
        f = filters.RegExpFilter("sleep", "root", "sleep")
        self.assertFalse(f.match([]))
        self.assertFalse(f.match(None))

    def test_empty_invalid_regexpfilter(self):
        f = filters.RegExpFilter("sleep", "root")
        self.assertFalse(f.match(["anything"]))
        self.assertFalse(f.match([]))

    def test_RegExpFilter_match(self):
        usercmd = ["ls", "/root"]
        filtermatch = wrapper.match_filter(self.filters, usercmd)
        self.assertFalse(filtermatch is None)
        self.assertEqual(
            ["/bin/ls", "/root"], filtermatch.get_command(usercmd)
        )

    def test_RegExpFilter_reject(self):
        usercmd = ["ls", "root"]
        self.assertRaises(
            wrapper.NoFilterMatched,
            wrapper.match_filter,
            self.filters,
            usercmd,
        )

    def test_missing_command(self):
        valid_but_missing = ["foo_bar_not_exist"]
        invalid = ["foo_bar_not_exist_and_not_matched"]
        self.assertRaises(
            wrapper.FilterMatchNotExecutable,
            wrapper.match_filter,
            self.filters,
            valid_but_missing,
        )
        self.assertRaises(
            wrapper.NoFilterMatched,
            wrapper.match_filter,
            self.filters,
            invalid,
        )

    def _test_EnvFilter_as_DnsMasq(self, config_file_arg):
        usercmd = [
            'env',
            config_file_arg + '=A',
            'NETWORK_ID=foobar',
            'dnsmasq',
            'foo',
        ]
        f = filters.EnvFilter(
            "env",
            "root",
            config_file_arg + '=A',
            'NETWORK_ID=',
            "/usr/bin/dnsmasq",
        )
        self.assertTrue(f.match(usercmd))
        self.assertEqual(['/usr/bin/dnsmasq', 'foo'], f.get_command(usercmd))
        env = f.get_environment(usercmd)
        assert env is not None
        self.assertEqual('A', env.get(config_file_arg))
        self.assertEqual('foobar', env.get('NETWORK_ID'))

    def test_EnvFilter(self):
        envset = ['A=/some/thing', 'B=somethingelse']
        envcmd = ['env'] + envset
        realcmd = ['sleep', '10']
        usercmd = envcmd + realcmd

        f = filters.EnvFilter("env", "root", "A=", "B=ignored", "sleep")
        # accept with leading env
        self.assertTrue(f.match(envcmd + ["sleep"]))
        # accept without leading env
        self.assertTrue(f.match(envset + ["sleep"]))

        # any other command does not match
        self.assertFalse(f.match(envcmd + ["sleep2"]))
        self.assertFalse(f.match(envset + ["sleep2"]))

        # accept any trailing arguments
        self.assertTrue(f.match(usercmd))

        # require given environment variables to match
        self.assertFalse(f.match(envcmd + ['C=ELSE']))
        self.assertFalse(f.match(['env', 'C=xx']))
        self.assertFalse(f.match(['env', 'A=xx']))

        # require env command to be given
        # (otherwise CommandFilters should match
        self.assertFalse(f.match(realcmd))
        # require command to match
        self.assertFalse(f.match(envcmd))
        self.assertFalse(f.match(envcmd[1:]))

        # ensure that the env command is stripped when executing
        self.assertEqual(realcmd, f.exec_args(usercmd))
        env = f.get_environment(usercmd)
        assert env is not None
        # check that environment variables are set
        self.assertEqual('/some/thing', env.get('A'))
        self.assertEqual('somethingelse', env.get('B'))
        self.assertNotIn('sleep', env.keys())

    def test_EnvFilter_without_leading_env(self):
        envset = ['A=/some/thing', 'B=somethingelse']
        envcmd = ['env'] + envset
        realcmd = ['sleep', '10']

        f = filters.EnvFilter("sleep", "root", "A=", "B=ignored")

        # accept without leading env
        self.assertTrue(f.match(envset + ["sleep"]))

        self.assertEqual(realcmd, f.get_command(envcmd + realcmd))
        self.assertEqual(realcmd, f.get_command(envset + realcmd))

        env = f.get_environment(envset + realcmd)
        assert env is not None
        # check that environment variables are set
        self.assertEqual('/some/thing', env.get('A'))
        self.assertEqual('somethingelse', env.get('B'))
        self.assertNotIn('sleep', env.keys())

    def test_KillFilter(self):
        if not os.path.exists(f"/proc/{os.getpid()}"):
            self.skipTest("Test requires /proc filesystem (procfs)")

        cat_cmd = "cat"
        cat_path = shutil.which(cat_cmd)
        if cat_path is None:
            self.skipTest("Test requires cat command")

        assert cat_path is not None
        # NOTE(tkajinam): In RHEL9 containers the cat command is a symlink
        #                 See bug 2037383 for details.
        if os.path.islink(cat_path):
            cat_path = os.path.realpath(cat_path)
            cat_cmd = os.path.basename(cat_path)

        p = subprocess.Popen(
            ["cat"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        try:
            f = filters.KillFilter("root", cat_path, "-9", "-HUP")
            usercmd = ['kill', '-ALRM', str(p.pid)]
            # Incorrect signal should fail
            self.assertFalse(f.match(usercmd))
            usercmd = ['kill', str(p.pid)]
            # Providing no signal should fail
            self.assertFalse(f.match(usercmd))
            # Providing matching signal should be allowed
            usercmd = ['kill', '-9', str(p.pid)]
            self.assertTrue(f.match(usercmd))

            f = filters.KillFilter("root", cat_path)
            usercmd = ['kill', str(os.getpid())]
            # Our own PID does not match /bin/sleep, so it should fail
            self.assertFalse(f.match(usercmd))
            usercmd = ['kill', '999999']
            # Nonexistent PID should fail
            self.assertFalse(f.match(usercmd))
            usercmd = ['kill', str(p.pid)]
            # Providing no signal should work
            self.assertTrue(f.match(usercmd))

            # verify that relative paths are matched against $PATH
            f = filters.KillFilter("root", cat_cmd)
            # Our own PID does not match so it should fail
            usercmd = ['kill', str(os.getpid())]
            self.assertFalse(f.match(usercmd))
            # Filter should find cat in /bin or /usr/bin
            usercmd = ['kill', str(p.pid)]
            self.assertTrue(f.match(usercmd))
            # Filter shouldn't be able to find binary in $PATH, so fail
            with fixtures.EnvironmentVariable("PATH", "/foo:/bar"):
                self.assertFalse(f.match(usercmd))
            # ensure that unset $PATH is not causing an exception
            with fixtures.EnvironmentVariable("PATH"):
                self.assertFalse(f.match(usercmd))
        finally:
            # Terminate the "cat" process and wait for it to finish
            p.terminate()
            p.wait()

    def test_KillFilter_no_raise(self):
        """Makes sure ValueError from bug 926412 is gone."""
        f = filters.KillFilter("root", "")
        # Providing anything other than kill should be False
        usercmd = ['notkill', '999999']
        self.assertFalse(f.match(usercmd))
        # Providing something that is not a pid should be False
        usercmd = ['kill', 'notapid']
        self.assertFalse(f.match(usercmd))
        # no arguments should also be fine
        self.assertFalse(f.match([]))
        self.assertFalse(f.match(None))

    def test_KillFilter_deleted_exe(self):
        """Makes sure deleted exe's are killed correctly."""
        command = "/bin/commandddddd"
        f = filters.KillFilter("root", command)
        usercmd = ['kill', '1234']
        # Providing no signal should work
        with mock.patch('os.readlink') as readlink:
            readlink.return_value = command + ' (deleted)'
            with mock.patch('os.path.isfile') as exists:

                def fake_exists(path):
                    return path == command

                exists.side_effect = fake_exists
                self.assertTrue(f.match(usercmd))

    @mock.patch('os.readlink')
    @mock.patch('os.path.isfile')
    def test_KillFilter_upgraded_exe(self, mock_isfile, mock_readlink):
        """Makes sure upgraded exe's are killed correctly."""
        f = filters.KillFilter("root", "/bin/commandddddd")
        command = "/bin/commandddddd"
        usercmd = ['kill', '1234']

        def fake_exists(path):
            return path == command

        mock_readlink.return_value = command + '\0\05190bfb2 (deleted)'
        mock_isfile.side_effect = fake_exists
        self.assertTrue(f.match(usercmd))

    @mock.patch('os.readlink')
    @mock.patch('os.path.isfile')
    @mock.patch('os.path.exists')
    @mock.patch('os.access')
    def test_KillFilter_renamed_exe(
        self, mock_access, mock_exists, mock_isfile, mock_readlink
    ):
        """Makes sure renamed exe's are killed correctly."""
        command = "/bin/commandddddd"
        f = filters.KillFilter("root", command)
        usercmd = ['kill', '1234']

        def fake_os_func(path, *args):
            return path == command

        mock_readlink.return_value = command + ';90bfb2 (deleted)'
        m = mock.mock_open(read_data=command)
        with mock.patch("builtins.open", m, create=True):
            mock_isfile.side_effect = fake_os_func
            mock_exists.side_effect = fake_os_func
            mock_access.side_effect = fake_os_func
            self.assertTrue(f.match(usercmd))

    def test_ReadFileFilter(self):
        goodfn = '/good/file.name'
        f = filters.ReadFileFilter(goodfn)
        usercmd = ['cat', '/bad/file']
        self.assertFalse(f.match(['cat', '/bad/file']))
        usercmd = ['cat', goodfn]
        self.assertEqual(['/bin/cat', goodfn], f.get_command(usercmd))
        self.assertTrue(f.match(usercmd))

    def test_IpFilter_non_netns(self):
        f = filters.IpFilter(self._ip, 'root')
        self.assertTrue(f.match(['ip', 'link', 'list']))
        self.assertTrue(f.match(['ip', '-s', 'link', 'list']))
        self.assertTrue(f.match(['ip', '-s', '-v', 'netns', 'add']))
        self.assertTrue(
            f.match(['ip', 'link', 'set', 'interface', 'netns', 'somens'])
        )

    def test_IpFilter_netns(self):
        f = filters.IpFilter(self._ip, 'root')
        self.assertFalse(f.match(['ip', 'netns', 'exec', 'foo']))
        self.assertFalse(f.match(['ip', 'netns', 'exec']))
        self.assertFalse(f.match(['ip', '-s', 'netns', 'exec']))
        self.assertFalse(f.match(['ip', '-l', '42', 'netns', 'exec']))
        self.assertFalse(f.match(['ip', 'net', 'exec', 'foo']))
        self.assertFalse(f.match(['ip', 'netns', 'e', 'foo']))

    def _test_IpFilter_netns_helper(self, action):
        f = filters.IpFilter(self._ip, 'root')
        self.assertTrue(f.match(['ip', 'link', action]))

    def test_IpFilter_netns_add(self):
        self._test_IpFilter_netns_helper('add')

    def test_IpFilter_netns_delete(self):
        self._test_IpFilter_netns_helper('delete')

    def test_IpFilter_netns_list(self):
        self._test_IpFilter_netns_helper('list')

    def test_IpNetnsExecFilter_match(self):
        f = filters.IpNetnsExecFilter(self._ip, 'root')
        self.assertTrue(
            f.match(['ip', 'netns', 'exec', 'foo', 'ip', 'link', 'list'])
        )
        self.assertTrue(f.match(['ip', 'net', 'exec', 'foo', 'bar']))
        self.assertTrue(f.match(['ip', 'netn', 'e', 'foo', 'bar']))
        self.assertTrue(f.match(['ip', 'net', 'e', 'foo', 'bar']))
        self.assertTrue(f.match(['ip', 'net', 'exe', 'foo', 'bar']))

    def test_IpNetnsExecFilter_nomatch(self):
        f = filters.IpNetnsExecFilter(self._ip, 'root')
        self.assertFalse(f.match(['ip', 'link', 'list']))
        self.assertFalse(f.match(['ip', 'foo', 'bar', 'netns']))
        self.assertFalse(f.match(['ip', '-s', 'netns', 'exec']))
        self.assertFalse(f.match(['ip', '-l', '42', 'netns', 'exec']))
        self.assertFalse(f.match(['ip', 'netns exec', 'foo', 'bar', 'baz']))
        self.assertFalse(f.match([]))

        # verify that at least a NS is given
        self.assertFalse(f.match(['ip', 'netns', 'exec']))

    def test_IpNetnsExecFilter_nomatch_nonroot(self):
        f = filters.IpNetnsExecFilter(self._ip, 'user')
        self.assertFalse(
            f.match(['ip', 'netns', 'exec', 'foo', 'ip', 'link', 'list'])
        )

    def test_match_filter_recurses_exec_command_filter_matches(self):
        filter_list = [
            filters.IpNetnsExecFilter(self._ip, 'root'),
            filters.IpFilter(self._ip, 'root'),
        ]
        args = ['ip', 'netns', 'exec', 'foo', 'ip', 'link', 'list']

        self.assertIsNotNone(wrapper.match_filter(filter_list, args))

    def test_match_filter_recurses_exec_command_matches_user(self):
        filter_list = [
            filters.IpNetnsExecFilter(self._ip, 'root'),
            filters.IpFilter(self._ip, 'user'),
        ]
        args = ['ip', 'netns', 'exec', 'foo', 'ip', 'link', 'list']

        # Currently ip netns exec requires root, so verify that
        # no non-root filter is matched, as that would escalate privileges
        self.assertRaises(
            wrapper.NoFilterMatched, wrapper.match_filter, filter_list, args
        )

    def test_match_filter_recurses_exec_command_filter_does_not_match(self):
        filter_list = [
            filters.IpNetnsExecFilter(self._ip, 'root'),
            filters.IpFilter(self._ip, 'root'),
        ]
        args = [
            'ip',
            'netns',
            'exec',
            'foo',
            'ip',
            'netns',
            'exec',
            'bar',
            'ip',
            'link',
            'list',
        ]

        self.assertRaises(
            wrapper.NoFilterMatched, wrapper.match_filter, filter_list, args
        )

    def test_ChainingRegExpFilter_match(self):
        filter_list = [
            filters.ChainingRegExpFilter('nice', 'root', 'nice', r'-?\d+'),
            filters.CommandFilter('cat', 'root'),
        ]
        args = ['nice', '5', 'cat', '/a']
        dirs = ['/bin', '/usr/bin']

        self.assertIsNotNone(wrapper.match_filter(filter_list, args, dirs))

    def test_ChainingRegExpFilter_not_match(self):
        filter_list = [
            filters.ChainingRegExpFilter('nice', 'root', 'nice', r'-?\d+'),
            filters.CommandFilter('cat', 'root'),
        ]
        args_invalid = (
            ['nice', '5', 'ls', '/a'],
            ['nice', '--5', 'cat', '/a'],
            ['nice2', '5', 'cat', '/a'],
            ['nice', 'cat', '/a'],
            ['nice', '5'],
        )
        dirs = ['/bin', '/usr/bin']

        for args in args_invalid:
            self.assertRaises(
                wrapper.NoFilterMatched,
                wrapper.match_filter,
                filter_list,
                args,
                dirs,
            )

    def test_ChainingRegExpFilter_multiple(self):
        filter_list = [
            filters.ChainingRegExpFilter(
                'ionice', 'root', 'ionice', '-c[0-3]'
            ),
            filters.ChainingRegExpFilter(
                'ionice', 'root', 'ionice', '-c[0-3]', '-n[0-7]'
            ),
            filters.CommandFilter('cat', 'root'),
        ]
        # both filters match to ['ionice', '-c2'], but only the second accepts
        args = ['ionice', '-c2', '-n7', 'cat', '/a']
        dirs = ['/bin', '/usr/bin']

        self.assertIsNotNone(wrapper.match_filter(filter_list, args, dirs))

    def test_ReadFileFilter_empty_args(self):
        goodfn = '/good/file.name'
        f = filters.ReadFileFilter(goodfn)
        self.assertFalse(f.match([]))
        self.assertFalse(f.match(None))

    def test_exec_dirs_search(self):
        # This test supposes you have /bin/cat or /usr/bin/cat locally
        f = filters.CommandFilter("cat", "root")
        usercmd = ['cat', '/f']
        self.assertTrue(f.match(usercmd))
        self.assertTrue(
            f.get_command(usercmd, exec_dirs=['/bin', '/usr/bin'])
            in (['/bin/cat', '/f'], ['/usr/bin/cat', '/f'])
        )

    def test_skips(self):
        # Check that all filters are skipped and that the last matches
        usercmd = ["cat", "/"]
        filtermatch = wrapper.match_filter(self.filters, usercmd)
        self.assertTrue(filtermatch is self.filters[-1])

    def test_RootwrapConfig(self):
        raw = configparser.RawConfigParser()

        # Empty config should raise configparser.Error
        self.assertRaises(configparser.Error, wrapper.RootwrapConfig, raw)

        # Check default values
        raw.set('DEFAULT', 'filters_path', '/a,/b')
        config = wrapper.RootwrapConfig(raw)
        self.assertEqual(['/a', '/b'], config.filters_path)
        self.assertEqual(os.environ["PATH"].split(':'), config.exec_dirs)

        with fixtures.EnvironmentVariable("PATH"):
            c = wrapper.RootwrapConfig(raw)
            self.assertEqual([], c.exec_dirs)

        self.assertFalse(config.use_syslog)
        self.assertEqual(
            logging.handlers.SysLogHandler.LOG_SYSLOG,
            config.syslog_log_facility,
        )
        self.assertEqual(logging.ERROR, config.syslog_log_level)

        # Check general values
        raw.set('DEFAULT', 'exec_dirs', '/a,/x')
        config = wrapper.RootwrapConfig(raw)
        self.assertEqual(['/a', '/x'], config.exec_dirs)

        raw.set('DEFAULT', 'use_syslog', 'oui')
        self.assertRaises(ValueError, wrapper.RootwrapConfig, raw)
        raw.set('DEFAULT', 'use_syslog', 'true')
        config = wrapper.RootwrapConfig(raw)
        self.assertTrue(config.use_syslog)

        raw.set('DEFAULT', 'syslog_log_facility', 'moo')
        self.assertRaises(ValueError, wrapper.RootwrapConfig, raw)
        raw.set('DEFAULT', 'syslog_log_facility', 'local0')
        config = wrapper.RootwrapConfig(raw)
        self.assertEqual(
            logging.handlers.SysLogHandler.LOG_LOCAL0,
            config.syslog_log_facility,
        )
        raw.set('DEFAULT', 'syslog_log_facility', 'LOG_AUTH')
        config = wrapper.RootwrapConfig(raw)
        self.assertEqual(
            logging.handlers.SysLogHandler.LOG_AUTH, config.syslog_log_facility
        )

        raw.set('DEFAULT', 'syslog_log_level', 'bar')
        self.assertRaises(ValueError, wrapper.RootwrapConfig, raw)
        raw.set('DEFAULT', 'syslog_log_level', 'INFO')
        config = wrapper.RootwrapConfig(raw)
        self.assertEqual(logging.INFO, config.syslog_log_level)

    def test_getlogin(self):
        with mock.patch('os.getlogin') as os_getlogin:
            os_getlogin.return_value = 'foo'
            self.assertEqual('foo', wrapper._getlogin())

    def test_getlogin_bad(self):
        with mock.patch('os.getenv') as os_getenv:
            with mock.patch('os.getlogin') as os_getlogin:
                os_getenv.side_effect = [None, None, 'bar']
                os_getlogin.side_effect = OSError(
                    '[Errno 22] Invalid argument'
                )
                self.assertEqual('bar', wrapper._getlogin())
                os_getlogin.assert_called_once_with()
                self.assertEqual(3, os_getenv.call_count)


class PathFilterTestCase(testtools.TestCase):
    def setUp(self):
        super().setUp()

        self.tmp_root_dir = tempfile.mkdtemp()
        tmpdir = fixtures.TempDir(self.tmp_root_dir)
        self.useFixture(tmpdir)

        self.f = filters.PathFilter('/bin/chown', 'root', 'nova', tmpdir.path)

        def gen_name():
            return str(uuid.uuid4())

        self.SIMPLE_FILE_WITHIN_DIR = os.path.join(tmpdir.path, 'some')
        self.SIMPLE_FILE_OUTSIDE_DIR = os.path.join(self.tmp_root_dir, 'some')
        self.TRAVERSAL_WITHIN_DIR = os.path.join(
            tmpdir.path, 'a', '..', 'some'
        )
        self.TRAVERSAL_OUTSIDE_DIR = os.path.join(tmpdir.path, '..', 'some')

        self.TRAVERSAL_SYMLINK_WITHIN_DIR = os.path.join(
            tmpdir.path, gen_name()
        )
        os.symlink(
            os.path.join(tmpdir.path, 'a', '..', 'a'),
            self.TRAVERSAL_SYMLINK_WITHIN_DIR,
        )

        self.TRAVERSAL_SYMLINK_OUTSIDE_DIR = os.path.join(
            tmpdir.path, gen_name()
        )
        os.symlink(
            os.path.join(tmpdir.path, 'a', '..', '..', '..', 'etc'),
            self.TRAVERSAL_SYMLINK_OUTSIDE_DIR,
        )

        self.SYMLINK_WITHIN_DIR = os.path.join(tmpdir.path, gen_name())
        os.symlink(os.path.join(tmpdir.path, 'a'), self.SYMLINK_WITHIN_DIR)

        self.SYMLINK_OUTSIDE_DIR = os.path.join(tmpdir.path, gen_name())
        os.symlink(
            os.path.join(self.tmp_root_dir, 'some_file'),
            self.SYMLINK_OUTSIDE_DIR,
        )

    def test_empty_args(self):
        self.assertFalse(self.f.match([]))
        self.assertFalse(self.f.match(None))

    def test_argument_pass_constraint(self):
        f = filters.PathFilter('/bin/chown', 'root', 'pass', 'pass')

        args = ['chown', 'something', self.SIMPLE_FILE_OUTSIDE_DIR]
        self.assertTrue(f.match(args))

    def test_argument_equality_constraint(self):
        temp_file_path = os.path.join(self.tmp_root_dir, 'spam/eggs')
        f = filters.PathFilter('/bin/chown', 'root', 'nova', temp_file_path)

        args = ['chown', 'nova', temp_file_path]
        self.assertTrue(f.match(args))

        args = ['chown', 'quantum', temp_file_path]
        self.assertFalse(f.match(args))

    def test_wrong_arguments_number(self):
        args = ['chown', '-c', 'nova', self.SIMPLE_FILE_WITHIN_DIR]
        self.assertFalse(self.f.match(args))

    def test_wrong_exec_command(self):
        args = ['wrong_exec', self.SIMPLE_FILE_WITHIN_DIR]
        self.assertFalse(self.f.match(args))

    def test_match(self):
        args = ['chown', 'nova', self.SIMPLE_FILE_WITHIN_DIR]
        self.assertTrue(self.f.match(args))

    def test_match_traversal(self):
        args = ['chown', 'nova', self.TRAVERSAL_WITHIN_DIR]
        self.assertTrue(self.f.match(args))

    def test_match_symlink(self):
        args = ['chown', 'nova', self.SYMLINK_WITHIN_DIR]
        self.assertTrue(self.f.match(args))

    def test_match_traversal_symlink(self):
        args = ['chown', 'nova', self.TRAVERSAL_SYMLINK_WITHIN_DIR]
        self.assertTrue(self.f.match(args))

    def test_reject(self):
        args = ['chown', 'nova', self.SIMPLE_FILE_OUTSIDE_DIR]
        self.assertFalse(self.f.match(args))

    def test_reject_traversal(self):
        args = ['chown', 'nova', self.TRAVERSAL_OUTSIDE_DIR]
        self.assertFalse(self.f.match(args))

    def test_reject_symlink(self):
        args = ['chown', 'nova', self.SYMLINK_OUTSIDE_DIR]
        self.assertFalse(self.f.match(args))

    def test_reject_traversal_symlink(self):
        args = ['chown', 'nova', self.TRAVERSAL_SYMLINK_OUTSIDE_DIR]
        self.assertFalse(self.f.match(args))

    def test_get_command(self):
        args = ['chown', 'nova', self.SIMPLE_FILE_WITHIN_DIR]
        expected = ['/bin/chown', 'nova', self.SIMPLE_FILE_WITHIN_DIR]

        self.assertEqual(expected, self.f.get_command(args))

    def test_get_command_traversal(self):
        args = ['chown', 'nova', self.TRAVERSAL_WITHIN_DIR]
        expected = [
            '/bin/chown',
            'nova',
            os.path.realpath(self.TRAVERSAL_WITHIN_DIR),
        ]

        self.assertEqual(expected, self.f.get_command(args))

    def test_get_command_symlink(self):
        args = ['chown', 'nova', self.SYMLINK_WITHIN_DIR]
        expected = [
            '/bin/chown',
            'nova',
            os.path.realpath(self.SYMLINK_WITHIN_DIR),
        ]

        self.assertEqual(expected, self.f.get_command(args))

    def test_get_command_traversal_symlink(self):
        args = ['chown', 'nova', self.TRAVERSAL_SYMLINK_WITHIN_DIR]
        expected = [
            '/bin/chown',
            'nova',
            os.path.realpath(self.TRAVERSAL_SYMLINK_WITHIN_DIR),
        ]

        self.assertEqual(expected, self.f.get_command(args))


class RunOneCommandTestCase(testtools.TestCase):
    def _test_returncode_helper(self, returncode: int, expected: int) -> None:
        with mock.patch.object(wrapper, 'start_subprocess') as mock_start:
            with mock.patch('sys.exit') as mock_exit:
                mock_start.return_value.wait.return_value = returncode
                # Using mocks for testing, so ignore arg types
                cmd.run_one_command(
                    None,  # type: ignore[arg-type]
                    mock.Mock(),
                    None,  # type: ignore[arg-type]
                    None,  # type: ignore[arg-type]
                )
        mock_exit.assert_called_once_with(expected)

    def test_positive_returncode(self):
        self._test_returncode_helper(1, 1)

    def test_negative_returncode(self):
        self._test_returncode_helper(-1, 129)


class DaemonCleanupException(Exception):
    pass


class DaemonCleanupTestCase(testtools.TestCase):
    @mock.patch('os.chmod')
    @mock.patch('shutil.rmtree')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch(
        'multiprocessing.managers.BaseManager.get_server',
        side_effect=DaemonCleanupException,
    )
    def test_daemon_no_cleanup_for_uninitialized_server(self, gs, mkd, *args):
        mkd.return_value = '/just_dir/123'
        self.assertRaises(
            DaemonCleanupException,
            daemon.daemon_start,
            config=None,
            filters=None,
        )



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

if os.environ.get('TEST_EVENTLET', False):
    import eventlet

    eventlet.monkey_patch()

    from oslo_rootwrap.tests import test_functional

    class RootwrapDaemonTest(test_functional.RootwrapDaemonTest):
        def assert_unpatched(self):
            # This test case is specifically for eventlet testing
            pass

        def _thread_worker(self, seconds, msg):
            code, out, err = self.execute(
                ['sh', '-c', f'sleep {seconds}; echo {msg}']
            )
            # Ignore trailing newline
            self.assertEqual(msg, out.rstrip())

        def _thread_worker_timeout(self, seconds, msg, timeout):
            with eventlet.Timeout(timeout):
                try:
                    self._thread_worker(seconds, msg)
                except eventlet.Timeout:
                    pass

        def test_eventlet_threads(self):
            """Check eventlet compatibility.

            The multiprocessing module is not eventlet friendly and
            must be protected against eventlet thread switching and its
            timeout exceptions.
            """
            th = []
            # 10 was not enough for some reason.
            for i in range(15):
                th.append(
                    eventlet.spawn(self._thread_worker, i % 3, f'abc{i}')
                )
            for i in [5, 17, 20, 25]:
                th.append(
                    eventlet.spawn(
                        self._thread_worker_timeout, 2, f'timeout{i}', i
                    )
                )
            for thread in th:
                thread.wait()



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import errno
import json
from multiprocessing import connection
from multiprocessing import managers
import socket
import struct
from typing import Any
import weakref

from oslo_rootwrap import wrapper


class RpcJSONEncoder(json.JSONEncoder):
    def default(self, o: object) -> Any:
        # We need to pass bytes unchanged as they are expected in arguments for
        # and are result of Popen.communicate()
        if isinstance(o, bytes):
            return {"__bytes__": base64.b64encode(o).decode('ascii')}
        # Handle two exception types relevant to command execution
        if isinstance(o, wrapper.NoFilterMatched):
            return {"__exception__": "NoFilterMatched"}
        elif isinstance(o, wrapper.FilterMatchNotExecutable):
            return {
                "__exception__": "FilterMatchNotExecutable",
                "match": o.match,
            }
        # Other errors will fail to pass JSON encoding and will be visible on
        # client side
        else:
            return super().default(o)


# Parse whatever RpcJSONEncoder supplied us with
def rpc_object_hook(obj: dict[str, Any]) -> Any:
    if "__exception__" in obj:
        type_name = obj.pop("__exception__")
        if type_name not in ("NoFilterMatched", "FilterMatchNotExecutable"):
            return obj
        exc_type = getattr(wrapper, type_name)
        return exc_type(**obj)
    elif "__bytes__" in obj:
        return base64.b64decode(obj["__bytes__"].encode('ascii'))
    else:
        return obj


class JsonListener:
    address: str
    closed: bool
    _socket: socket.socket
    _accepted: weakref.WeakSet["JsonConnection"]

    def __init__(self, address: str, backlog: int = 1) -> None:
        self.address = address
        self._socket = socket.socket(socket.AF_UNIX)
        try:
            self._socket.setblocking(True)
            self._socket.bind(address)
            self._socket.listen(backlog)
        except OSError:
            self._socket.close()
            raise
        self.closed = False
        self._accepted = weakref.WeakSet()

    def accept(self) -> "JsonConnection":
        while True:
            try:
                s, _ = self._socket.accept()
            except OSError as e:
                if e.errno in (errno.EINVAL, errno.EBADF):
                    raise EOFError
                elif e.errno != errno.EINTR:
                    raise
            else:
                break
        s.setblocking(True)
        conn = JsonConnection(s)
        self._accepted.add(conn)
        return conn

    def close(self) -> None:
        if not self.closed:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
            self.closed = True

    def get_accepted(self) -> weakref.WeakSet["JsonConnection"]:
        return self._accepted


if hasattr(managers.Server, 'accepter'):
    # In Python 3 accepter() thread has infinite loop. We break it with
    # EOFError, so we should silence this error here.
    def silent_accepter(self: managers.Server) -> None:
        try:
            old_accepter(self)
        except EOFError:
            pass

    old_accepter = managers.Server.accepter
    managers.Server.accepter = silent_accepter  # type: ignore[method-assign]


class JsonConnection:
    _socket: socket.socket

    def __init__(self, sock: socket.socket) -> None:
        sock.setblocking(True)
        self._socket = sock

    def send_bytes(self, s: bytes) -> None:
        self._socket.sendall(struct.pack('!Q', len(s)))
        self._socket.sendall(s)

    def recv_bytes(self, maxsize: int | None = None) -> bytes:
        item = struct.unpack('!Q', self.recvall(8))[0]
        if maxsize is not None and item > maxsize:
            raise RuntimeError("Too big message received")
        s = self.recvall(item)
        return s

    def send(self, obj: object) -> None:
        s = self.dumps(obj)
        self.send_bytes(s)

    def recv(self) -> Any:
        s = self.recv_bytes()
        return self.loads(s)

    def close(self) -> None:
        self._socket.close()

    def half_close(self) -> None:
        self._socket.shutdown(socket.SHUT_RD)

    # We have to use slow version of recvall with eventlet because of a bug in
    # GreenSocket.recv_into:
    # https://bitbucket.org/eventlet/eventlet/pull-request/41
    def _recvall_slow(self, size: int) -> bytes:
        remaining = size
        res: list[bytes] = []
        while remaining:
            piece = self._socket.recv(remaining)
            if not piece:
                raise EOFError
            res.append(piece)
            remaining -= len(piece)
        return b''.join(res)

    def recvall(self, size: int) -> bytes:
        buf = bytearray(size)
        mem = memoryview(buf)
        got = 0
        while got < size:
            piece_size = self._socket.recv_into(mem[got:])
            if not piece_size:
                raise EOFError
            got += piece_size
        # bytearray is mostly compatible with bytes and we could avoid copying
        # data here, but hmac doesn't like it in Python 3.3 (not in 2.7 or 3.4)
        return bytes(buf)

    @staticmethod
    def dumps(obj: object) -> bytes:
        return json.dumps(obj, cls=RpcJSONEncoder).encode('utf-8')

    @staticmethod
    def loads(s: bytes) -> Any:
        res = json.loads(s.decode('utf-8'), object_hook=rpc_object_hook)
        try:
            kind = res[0]
        except (IndexError, TypeError):
            pass
        else:
            # In Python 2 json returns unicode while multiprocessing needs str
            if kind in ("#TRACEBACK", "#UNSERIALIZABLE") and not isinstance(
                res[1], str
            ):
                res[1] = res[1].encode('utf-8', 'replace')
        return res


class JsonClient(JsonConnection):
    def __init__(self, address: str, authkey: bytes | None = None) -> None:
        sock = socket.socket(socket.AF_UNIX)
        sock.setblocking(True)
        sock.connect(address)
        super().__init__(sock)
        if authkey is not None:
            connection.answer_challenge(self, authkey)  # type: ignore[arg-type]
            connection.deliver_challenge(self, authkey)  # type: ignore[arg-type]



# Copyright (c) 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import sys
import threading

from oslo_rootwrap import cmd
from oslo_rootwrap import subprocess


def forward_stream(fr, to):
    while True:
        line = fr.readline()
        if not line:
            break
        to.write(line)


def forwarding_popen(f, old_popen=subprocess.Popen):
    def popen(*args, **kwargs):
        p = old_popen(*args, **kwargs)
        t = threading.Thread(target=forward_stream, args=(p.stderr, f))
        t.daemon = True
        t.start()
        return p

    return popen


class nonclosing:
    def __init__(self, f):
        self._f = f

    def __getattr__(self, name):
        return getattr(self._f, name)

    def close(self):
        pass


log_format = "%(asctime)s | [%(process)5s]+%(levelname)5s | %(message)s"
if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, format=log_format)
    sys.stderr = nonclosing(sys.stderr)
    cmd.daemon()



# Copyright (c) 2015 Red Hat.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import debtcollector

__all__ = [
    'subprocess',
]

try:
    import eventlet.patcher
except ImportError:
    _patched_socket = False
else:
    # In tests patching happens later, so we'll rely on environment variable
    _patched_socket = eventlet.patcher.is_monkey_patched('socket') or bool(
        os.environ.get('TEST_EVENTLET', False)
    )

if not _patched_socket:
    import subprocess
else:
    debtcollector.deprecate(
        "Eventlet support is deprecated and will be soon removed."
    )
    from eventlet.green import subprocess  # type: ignore
